#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Tuple, List, Union

import numpy as np

from piquasso.api.config import Config
from piquasso.api.exceptions import InvalidState, InvalidParameter, PiquassoException
from piquasso.api.state import State
from piquasso.api.connector import BaseConnector

from piquasso._math.functions import gaussian_wigner_function
from piquasso._math.linalg import (
    is_symmetric,
    is_positive_semidefinite,
)
from piquasso._math.symplectic import symplectic_form, xp_symplectic_form
from piquasso._math.fock import get_fock_space_basis
from piquasso._math.transformations import xxpp_to_xpxp_indices, xpxp_to_xxpp_indices

from piquasso._math.decompositions import williamson
from piquasso.api.exceptions import InvalidModes
from piquasso._simulators.plot import plot_wigner_function

from .probabilities import (
    DensityMatrixCalculation,
    DisplacedDensityMatrixCalculation,
    NondisplacedDensityMatrixCalculation,
    calculate_click_probability_nondisplaced,
    calculate_click_probability,
)


class GaussianState(State):
    r"""Class to represent a Gaussian state."""

    def __init__(
        self, d: int, connector: BaseConnector, config: Optional[Config] = None
    ) -> None:
        """
        Args:
            d (int): The number of modes.

        """
        super().__init__(connector=connector, config=config)
        self._d = d
        self.reset()

    def __len__(self) -> int:
        return self._d

    @property
    def d(self) -> int:
        return len(self)

    def reset(self) -> None:
        r"""Resets the state to a vacuum."""

        np = self._connector.np

        vector_shape = (self.d,)
        matrix_shape = vector_shape * 2

        self._m = np.zeros(vector_shape, dtype=self._config.complex_dtype)
        self._G = np.zeros(matrix_shape, dtype=self._config.complex_dtype)
        self._C = np.zeros(matrix_shape, dtype=self._config.complex_dtype)

    @classmethod
    def _from_representation(
        cls,
        *,
        m: np.ndarray,
        G: np.ndarray,
        C: np.ndarray,
        config: Config,
        connector: BaseConnector,
    ) -> "GaussianState":
        obj = cls(d=len(m), connector=connector, config=config)

        obj._m = m
        obj._G = G
        obj._C = C

        return obj

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, GaussianState):
            return False

        return (
            np.allclose(self._C, other._C)
            and np.allclose(self._G, other._G)
            and np.allclose(self._m, other._m)
        )

    def validate(self) -> None:
        r"""
        Validates the state.

        Raises:
            InvalidState:
                Raised if the underlying Gaussian state is invalid, which could mean
                - ill-shaped mean and covariance;
                - non-symmetric covariance matrix;
                - the covariance matrix doesn't fulfill the Robertson-Schrödinger
                  uncertainty relations.
        """
        self._validate_mean(self.xpxp_mean_vector, self.d)
        self._validate_cov(self.xpxp_covariance_matrix, self.d)

    def _validate_mean(self, mean: np.ndarray, d: int) -> None:
        if not self._config.validate:
            return

        expected_shape = (2 * d,)

        if not mean.shape == expected_shape:
            raise InvalidState(
                f"Invalid 'mean' vector shape; "
                f"expected={expected_shape}, actual={mean.shape}."
            )

    def _validate_cov(self, cov: np.ndarray, d: int) -> None:
        if not self._config.validate:
            return

        expected_shape = (2 * d,) * 2

        if not cov.shape == expected_shape:
            raise InvalidState(
                f"Invalid 'cov' matrix shape; "
                f"expected={expected_shape}, actual={cov.shape}."
            )

        if not is_symmetric(cov):
            raise InvalidState("The covariance matrix is not symmetric.")

        if not is_positive_semidefinite(
            cov / self._config.hbar + 1j * symplectic_form(d)
        ):
            raise InvalidState(
                "The covariance matrix is invalid, since it doesn't fulfill the "
                "Robertson-Schrödinger uncertainty relation."
            )

    @property
    def xxpp_mean_vector(self) -> np.ndarray:
        r"""The state's mean in the xxpp-ordered basis.

        The expectation value of the quadrature operators in xxpp basis, i.e.
        :math:`\operatorname{Tr} \left ( \rho Y \right )`, where
        :math:`Y = (x_1, \dots, x_d, p_1, \dots, p_d)^T`.

        Returns:
            numpy.ndarray: A :math:`2d`-vector where :math:`d` is the number of modes.
        """

        np = self._connector.np

        dimensionless_xxpp_mean_vector = np.concatenate(
            [self._m.real, self._m.imag]
        ) * np.sqrt(2)

        return dimensionless_xxpp_mean_vector * np.sqrt(self._config.hbar)

    @xxpp_mean_vector.setter
    def xxpp_mean_vector(self, value):
        self.xpxp_mean_vector = value[xxpp_to_xpxp_indices(self.d)]

    @property
    def xxpp_covariance_matrix(self) -> np.ndarray:
        r"""The xxpp-ordered covariance matrix of the state.

        The xxpp-ordered covariance matrix :math:`\sigma_{xp}` is defined by

        .. math::
            \sigma_{xp, ij} := \langle Y_i Y_j + Y_j Y_i \rangle_\rho
                - 2 \langle Y_i \rangle_\rho \langle Y_j \rangle_\rho,

        where

        .. math::
            Y = (x_1, \dots, x_d, p_1, \dots, p_d)^T,

        and :math:`\rho` is the density operator of the currently represented state.

        Returns:
            numpy.ndarray:
                The :math:`2d \times 2d` xp-ordered covariance matrix.
        """
        np = self._connector.np

        C = self._C
        G = self._G

        dimensionless_xxpp_covariance_matrix = 2 * np.block(
            [
                [(G + C).real, (G + C).imag],
                [(G - C).imag, (-G + C).real],
            ],
        ) + np.identity(2 * self.d)

        return dimensionless_xxpp_covariance_matrix * self._config.hbar

    @xxpp_covariance_matrix.setter
    def xxpp_covariance_matrix(self, value):
        indices = xxpp_to_xpxp_indices(self.d)
        self.xpxp_covariance_matrix = value[np.ix_(indices, indices)]

    @property
    def xxpp_correlation_matrix(self) -> np.ndarray:
        r"""The state's correlation matrix in the xxpp basis.

        Let :math:`M_{(xxpp)}` be the correlation matrix in the xxpp basis.
        Then

        .. math::
            M_{ij (xxpp)} = \langle Y_i Y_j + Y_j Y_i \rangle_\rho,

        where :math:`M_{ij (xxpp)}` denotes the matrix element in the
        :math:`i`-th row and :math:`j`-th column,

        .. math::
            Y = (x_1, \dots, x_d, p_1, \dots, p_d)^T,

        and :math:`\rho` is the density operator of the currently represented state.

        Returns:
            numpy.ndarray: The :math:`2d \times 2d` correlation matrix in the
            xxpp-basis.
        """
        np = self._connector.np

        xxpp_mean_vector = self.xxpp_mean_vector

        return self.xxpp_covariance_matrix + 2 * np.outer(
            xxpp_mean_vector, xxpp_mean_vector
        )

    @property
    def xxpp_representation(self) -> Tuple[np.ndarray, np.ndarray]:
        r"""
        The state's mean and correlation matrix ordered in the xxpp basis.

        Returns:
            tuple: :meth:`xxpp_mean_vector`, :meth:`xxpp_correlation_matrix`.
        """

        return self.xxpp_mean_vector, self.xxpp_correlation_matrix

    @property
    def xpxp_mean_vector(self) -> np.ndarray:
        r"""Returns the xpxp-ordered mean of the state.

        Returns:
            numpy.ndarray: A :math:`2d`-vector. The expectation value of the quadrature
            operators in `xxpp`-ordering, i.e. :math:`\operatorname{Tr} \rho R`, where
            :math:`R = (x_1, p_1, \dots, x_d, p_d)^T`.
        """
        return self.xxpp_mean_vector[xxpp_to_xpxp_indices(self.d)]

    @xpxp_mean_vector.setter
    def xpxp_mean_vector(self, value: np.ndarray) -> None:
        np = self._connector.np

        self._validate_mean(value, self.d)

        m = (value[::2] + 1j * value[1::2]) / np.sqrt(2 * self._config.hbar)

        self._m = m

    @property
    def xpxp_covariance_matrix(self) -> np.ndarray:
        r"""The `xpxp`-ordered covariance matrix of the state.

        The `xpxp`-ordered covariance matrix :math:`\sigma` is defined by

        .. math::
            \sigma_{ij} := \langle R_i R_j + R_j R_i \rangle_\rho
                - 2 \langle R_i \rangle_\rho \langle R_j \rangle_\rho,

        where

        .. math::
            R = (x_1, p_1, \dots, x_d, p_d)^T,

        and :math:`\rho` is the density operator of the currently represented state.

        Returns:
            numpy.ndarray:
                The :math:`2d \times 2d` `xpxp`-ordered covariance matrix in
                xpxp-ordered basis.
        """

        indices = xxpp_to_xpxp_indices(self.d)
        return self.xxpp_covariance_matrix[np.ix_(indices, indices)]

    @xpxp_covariance_matrix.setter
    def xpxp_covariance_matrix(self, new_cov: np.ndarray) -> None:
        np = self._connector.np

        d = self.d

        self._validate_cov(new_cov, d)

        indices = xpxp_to_xxpp_indices(self.d)

        dimensionless_cov = new_cov / self._config.hbar
        dimensionless_xp_cov = dimensionless_cov[np.ix_(indices, indices)]

        blocks = (dimensionless_xp_cov - np.identity(2 * d)) / 4

        C_real = blocks[:d, :d] + blocks[d:, d:]
        G_real = blocks[:d, :d] - blocks[d:, d:]

        C_imag = blocks[:d, d:] - blocks[d:, :d]
        G_imag = blocks[:d, d:] + blocks[d:, :d]

        C = C_real + 1j * C_imag
        G = G_real + 1j * G_imag

        self._G = G
        self._C = C

    @property
    def xpxp_correlation_matrix(self) -> np.ndarray:
        r"""The `xpxp`-ordered correlation matrix of the state.

        Let :math:`M` be the correlation matrix in the `xpxp` basis.
        Then

        .. math::
            M_{ij} = \langle R_i R_j + R_j R_i \rangle_\rho,

        where :math:`M_{ij}` denotes the matrix element in the
        :math:`i`-th row and :math:`j`-th column,

        .. math::
            R = (x_1, p_1, \dots, x_d, p_d)^T,

        and :math:`\rho` is the density operator of the currently represented state.

        Returns:
            numpy.ndarray: The :math:`2d \times 2d` `xpxp`-ordered correlation matrix.
        """

        indices = xxpp_to_xpxp_indices(self.d)
        return self.xxpp_correlation_matrix[np.ix_(indices, indices)]

    @property
    def xpxp_representation(self) -> Tuple[np.ndarray, np.ndarray]:
        r"""The state's mean and correlation matrix ordered by the `xpxp` basis.

        Returns:
            tuple: :meth:`mean`, :meth:`corr`.
        """

        return self.xpxp_mean_vector, self.xpxp_correlation_matrix

    @property
    def complex_displacement(self) -> np.ndarray:
        r"""The complex displacement of the state.

        The complex displacement is defined by

        .. math::
            \mu_{c} := \begin{bmatrix}
                \langle a_1 \rangle_{\rho}, \dots, \langle a_d \rangle_{\rho},
                \langle a^\dagger_1 \rangle_{\rho},
                \dots, \langle a^\dagger_d \rangle_{\rho}
            \end{bmatrix}.

        Equivalently, one can write

        .. math::
            \mu_{c} := W \mu_{xxpp},

        where :math:`\mu_{xxpp}` is the xxpp-ordered mean vector
        :attr:`xxpp_mean_vector` and

        .. math::
            W = \frac{1}{\sqrt{2}} \begin{bmatrix}
                I_{d} & i I_{d} \\
                I_{d} & -i I_{d}
            \end{bmatrix}.

        Returns:
            numpy.ndarray: The complex displacement.
        """
        np = self._connector.np

        return np.concatenate([self._m, self._m.conj()])

    @property
    def complex_covariance(self) -> np.ndarray:
        r"""The complex covariance of the state.

        The complex covariance is defined by

        .. math::
            \sigma_{c, ij} = \langle
                \xi_i \xi^\dagger_j + \xi_j \xi^\dagger_i
            \rangle_{\rho}
            - 2 \langle \xi_i \rangle_{\rho} \langle \xi^\dagger_j \rangle_{\rho},

        where

        .. math::
            \xi = \begin{bmatrix}
                a_1, \dots a_d, a^\dagger_1, \dots, a^\dagger_d
            \end{bmatrix}.

        Equivalently, one can write

        .. math::
            \sigma_{c} = \frac{1}{\hbar} W \sigma_{xxpp} W^{\dagger},

        where :math:`\sigma_{xxpp}` is the xxpp-ordered covariance
        matrix :attr:`xxpp_cov` and

        .. math::
            W = \frac{1}{\sqrt{2}} \begin{bmatrix}
                I_{d} & i I_{d} \\
                I_{d} & -i I_{d}
            \end{bmatrix}.

        Returns:
            numpy.ndarray: The complex covariance.
        """

        np = self._connector.np

        return 2 * np.block(
            [[self._C.conj(), self._G], [self._G.conj(), self._C]]
        ) + np.identity(2 * self.d)

    @property
    def Q_matrix(self):
        return (self.complex_covariance + np.identity(2 * len(self))) / 2

    def rotated(self, phi: float) -> "GaussianState":
        r"""Returns a copy of the current Gaussian state, rotated by an angle `phi`.

        Let :math:`\phi \in [ 0, 2 \pi )`. Let us define the following:

        .. math::
            x_{i, \phi} = \cos\phi~x_i + \sin\phi~p_i,

        which is a generalized quadrature operator. One could rotate the state by a
        simple complex phase, which can be shown by using the transformation rules
        between the ladder operators and quadrature operators, i.e.

        .. math::
            x_i &= \sqrt{\frac{\hbar}{2}} (a_i + a_i^\dagger) \\
            p_i &= -i \sqrt{\frac{\hbar}{2}} (a_i - a_i^\dagger),

        which we could rewrite :math:`x_{i, \phi}` to

        .. math::
            x_{i, \phi} = \sqrt{\frac{\hbar}{2}} \left(
                a_i \exp(-i \phi) + a_i^\dagger \exp(i \phi)
            \right),

        meaning that e.g. the annihilation operators :math:`a_i` are transformed just
        by multiplying it with a phase factor :math:`\exp(-i \phi)` i.e.

        .. math::
            (\langle a_i \rangle_{\rho} =: )~m_i &\mapsto \exp(-i \phi) m_i \\
            (\langle a^\dagger_i a_j \rangle_{\rho} =: )~C_{ij} &\mapsto C_{ij} \\
            (\langle a_i a_j \rangle_{\rho} =: )~G_{ij} &\mapsto \exp(-i 2 \phi) G_{ij}.

        Args:
            phi (float): The angle to rotate the state with.

        Returns:
            GaussianState: The rotated `GaussianState` instance.
        """
        np = self._connector.np

        phase = np.exp(-1j * phi)

        return self.__class__._from_representation(
            C=self._C,
            G=(self._G * phase**2),
            m=(self._m * phase),
            config=self._config,
            connector=self._connector,
        )

    def reduced(self, modes: Tuple[int, ...]) -> "GaussianState":
        """Returns a copy of the current state, reduced to the given `modes`.

        This method essentially preserves the modes specified from the representation
        of the Gaussian state, but cuts out the other modes.

        Args:
            modes (tuple[int]): The modes to reduce the state to.

        Returns:
            GaussianState: The reduced `GaussianState` instance.
        """
        return self.__class__._from_representation(
            C=self._C[np.ix_(modes, modes)],
            G=self._G[np.ix_(modes, modes)],
            m=self._m[np.ix_(modes)],
            config=self._config,
            connector=self._connector,
        )

    def xpxp_reduced_rotated_mean_and_covariance(
        self, modes: Tuple[int, ...], phi: float
    ) -> Tuple[np.ndarray, np.ndarray]:
        r"""The mean and covariance on a rotated and reduced state.

        Let the index set :math:`\vec{i}` correspond to `modes`, and the angle
        :math:`\phi` correspond to `phi`. The current :class:`GaussianState` instance
        is reduced to `modes` and rotated by `phi` in a new instance, and let that
        state be denoted by :math:`\rho_{\vec{i}, \phi}`.

        Then the `xpxp`-ordered mean and covariance can be calculated by

        .. math::
            \mu_{\vec{i}, \phi}
                &:= \langle R_{\vec{i}} \rangle_{\rho_{\vec{i}, \phi}}, \\
            \sigma_{\vec{i}, \phi}
                &:=  \langle
                    R_{\vec{i}} R_{\vec{i}}^T
                \rangle_{\rho_{\vec{i}, \phi}}
                - \mu_{\vec{i}, \phi} \mu_{\vec{i}, \phi}^T,

        where

        .. math::
            R = (x_1, p_1, \dots, x_d, p_d)^T,

        and :math:`R_{\vec{i}}` is just the same vector, reduced to a subsystem
        specified by :math:`\vec{i}`.

        Args:
            modes (tuple[int]): The modes to reduce the state to.
            phi (float): The angle to rotate the state with.

        Returns:
            (numpy.ndarray, numpy.ndarray):
                `xpxp`-ordered mean and covariance of the reduced and rotated
                version of the current :class:`GaussianState`.
        """
        transformed_state = self.reduced(modes).rotated(phi)

        return (
            transformed_state.xpxp_mean_vector,
            transformed_state.xpxp_covariance_matrix,
        )

    def mean_photon_number(self, modes: Optional[Tuple[int, ...]] = None) -> float:
        r"""This method returns the mean photon number of the given modes.
        The mean photon number :math:`\bar{n} = \langle \hat{n}
        \rangle` can be calculated in terms of the ladder operators by the following
        expression

        .. math::
            \sum_{i=1}^{d} \langle a_{i}^\dagger a_{i} \rangle =
            \operatorname{Tr}(\rho \hat{n}),

        where :math:`a`, :math:`a ^\dagger` are the annihilation and the creation
        operators respectively, :math:`\rho` is the density operator of the
        currently represented state and :math:`d` is the number of modes. For a general
        displaced squeezed gaussian state, the mean photon number is

        .. math::
            \langle \hat{n} \rangle = \operatorname{Tr}(\langle a^\dagger a \rangle) +
            \mu_{c}^ \dagger \cdot \mu_{c},

        where :math:`\mu_{c}` is the :attr:`complex_displacement`.

        .. note::
            This method can also be used to return the summation of the mean photon
            number for multiple modes if the mode parameter contains more than one
            integer e.g :math:`(0,1,...)`.

        Args:
            modes (tuple[int]): The correspoding modes at which the mean photon number
                is calculated.
        Returns:
            float: The expectation value of the photon number.
        """

        if modes is not None:
            return self.reduced(modes).mean_photon_number()

        np = self._connector.np

        return (np.trace(self._C) + self._m.conjugate() @ self._m).real

    def variance_photon_number(self, modes: Optional[Tuple[int, ...]] = None) -> float:
        r"""
        This method calculates the variance of the photon number operator as follows:

        .. math::
            \operatorname{Var}(\hat{n}) = \mathbb{E}((\hat{n} - \bar{n})^2),

        where :math:`\bar{n}` is the expectation value of the photon number, given by
        :meth:`mean_photon_number`.

        See, e.g., [Means and covariances of photon numbers in multimode Gaussian states](https://journals.aps.org/pra/abstract/10.1103/PhysRevA.99.023817)
        for details on how to calculate this for Gaussian states.

        Args:
            modes (Tuple[int, ...]): The correspoding modes at which the variance of
                the photon number is calculated.

        Returns:
            float: Variance of the photon number operator
        """  # noqa: E501

        if modes is not None:
            return self.reduced(modes).variance_photon_number()

        np = self._connector.np

        d = self._d

        m = self._m

        m_outer = np.outer(m, m)

        G = self._G + m_outer
        C = self._C + np.outer(np.conj(m), m)

        correlation_term = (
            np.einsum("kj,jk", np.conj(G), G)
            + np.einsum("jk,kj", C, np.identity(d) + C)
            + np.einsum("jj,kk", C, C)
            - 2 * np.sum(np.abs(m_outer) ** 2)
        )

        mean = self.mean_photon_number()

        variance = (correlation_term - mean**2).real

        return variance

    def fidelity(self, state: "GaussianState") -> float:
        r"""Calculates the state fidelity between two quantum states.

        The state fidelity :math:`F` between two density matrices
        :math:`\rho_1, \rho_2` is given by:

        .. math::
            \operatorname{F}(\rho_1, \rho_2) = \left [
                \operatorname{Tr} \sqrt{\sqrt{\rho_1}
                \rho_2 \sqrt{\rho_1} }
            \right ]

        A Gaussian state can be represented by its covariance matrix and displacement
        vector.

        Let :math:`\mu_1, \mu_2` be the displacement vectors and
        :math:`\sigma_1, \sigma_2` be the covariance matrices of the
        :math:`\rho_1, \rho_2` Gaussian states, respectively. Define
        :math:`\hat{\sigma} = \frac{\sigma_1 + \sigma_2}{2}` and
        :math:`\Delta\mu = \mu_2 - \mu_1`.
        The fidelity can be written as

        .. math::
            \operatorname{F}(\rho_1, \rho_2) = \operatorname{F_0}(\sigma_1, \sigma_2)
            \exp \left(
                -\frac{1}{4} \Delta\mu^T (\hat{\sigma})^{-1} \Delta\mu
            \right),

        where :math:`F_0` can be written as

        .. math::
            \operatorname{F_0} = \frac{
                \prod_{i=1}^d \left [w_i + \sqrt{w_i^2 - 1} \right]
            }{
                \sqrt{\det \hat{\sigma}}
            },

        where :math:`w_i \geq 1` and :math:`\pm w_i` are the eigenvalues of the matrix

        .. math::
            W := - \frac{i}{2} \Omega^T \hat{\sigma}^{-1} \left(
                I - \sigma_2 \Omega \sigma_1 \Omega
            \right)

        and

        .. math::
            \Omega = \begin{bmatrix}
                0 & 1 \\-1 & 0
            \end{bmatrix} \otimes I_{2d \times 2d}.

        References:
            - `Quantum fidelity for arbitrary Gaussian states <https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.115.260501>`_.

        Note:
            In this notation :math:`\sqrt{\hbar} \mu_i` is equivalent to
            :attr:`xpxp_mean_vector` and :math:`\hbar \sigma_i` is equivalent to
            :attr:`xpxp_covariance_matrix`.

        Args:
            state: Another :class:`~piquasso._simulators.gaussian.state.GaussianState`
                instance.

        Returns:
            float: The calculated fidelity.
        """  # noqa: E501

        np = self._connector.np

        hbar = self._config.hbar

        sigma_1 = self.xpxp_covariance_matrix / hbar
        sigma_2 = state.xpxp_covariance_matrix / hbar

        sigma_mean = (sigma_1 + sigma_2) / 2

        Omega = symplectic_form(self.d)

        Id = np.identity(2 * self.d)

        W_aux = (
            -1j
            / 2
            * Omega.T
            @ np.linalg.inv(sigma_mean)
            @ (Id - sigma_2 @ Omega @ sigma_1 @ Omega)
        )

        eigenvalues = np.linalg.eigvals(W_aux)
        positive_eigenvalues = eigenvalues[eigenvalues >= 0]

        F_0 = np.prod(
            [w + np.sqrt(w**2 - 1) for w in positive_eigenvalues]
        ) / np.sqrt(np.linalg.det(sigma_mean))

        mu_1 = self.xpxp_mean_vector / np.sqrt(hbar)
        mu_2 = state.xpxp_mean_vector / np.sqrt(hbar)
        delta_mu = mu_2 - mu_1
        displaced_contribition = np.exp(
            -1 / 2 * delta_mu @ np.linalg.inv(sigma_mean) @ delta_mu
        )

        return np.real(displaced_contribition * F_0)

    def quadratic_polynomial_expectation(
        self, A: np.ndarray, b: np.ndarray, c: float = 0.0, phi: float = 0.0
    ) -> float:
        r"""The expectation value of the specified quadratic polynomial.

        A quadratic polynomial can be written as

        .. math::
            f(R) = R^T A R + R \cdot b + c,

        where :math:`R = (x_1, p_1, \dots, x_d, p_d)^T` is the vector of the quadrature
        operators where :math:`d` is the number of modes,
        :math:`A \in \mathbb{R}^{2d \times 2d}` is a symmetric matrix,
        :math:`b \in \mathbb{R}^{2d}`, and :math:`c\in\mathbb{R}`.

        This method returns the expectation value :math:`E[f(R)]` using the following
        equation:

        .. math::
            \operatorname{E}[f(R)]
                = \operatorname{Tr}[ A\sigma ] + \mu^T A \mu + \mu^T b + c,

        where :math:`\sigma` is the covariance matrix, :math:`\mu = E[R]` is the mean
        of the quadrature operators and

        .. math::
            \operatorname{E}[\cdot] = \operatorname{Tr}[\cdot \rho],

        where :math:`\rho` is the density matrix of the state.

        Args:
            A (numpy.ndarray):
                A :math:`2d \times 2d` real symmetric matrix corresponding to the
                quadratic coefficients, where :math:`d` is the number of modes.
            b (numpy.ndarray):
                A one-dimensional :math:`2d`-length real-valued vector that corresponds
                to the first order terms of the quadratic polynomial.
            c (float): The constant term in the quadratic polynomial. Defaults to `0`.
            phi (float): Rotation angle, by which the state is rotated. Defaults to `0`.
        Returns:
            float: The expectation value of the quadratic polynomial.
        """

        np = self._connector.np

        if self._config.validate and not is_symmetric(A):
            raise InvalidParameter("The specified matrix is not symmetric.")

        state = self.rotated(phi)
        mean = state.xpxp_mean_vector
        cov = state.xpxp_covariance_matrix
        first_moment = np.trace(A @ cov) / 2 + mean @ A @ mean + mean @ b + c
        # TODO: calculate the variance.
        return first_moment

    def _is_displaced(self) -> bool:
        return not self._connector.np.allclose(self._m, 0.0)

    def _get_density_matrix_calculation(self) -> DensityMatrixCalculation:
        if self._connector.allow_conditionals and not self._is_displaced():
            return NondisplacedDensityMatrixCalculation(
                complex_covariance=self.complex_covariance, connector=self._connector
            )
        return DisplacedDensityMatrixCalculation(
            complex_displacement=self.complex_displacement,
            complex_covariance=self.complex_covariance,
            connector=self._connector,
        )

    @property
    def density_matrix(self) -> np.ndarray:
        """The density matrix of the state in the truncated Fock space."""
        calculation = self._get_density_matrix_calculation()

        return calculation.get_density_matrix(
            get_fock_space_basis(d=self.d, cutoff=self._config.cutoff)
        )

    def wigner_function(
        self,
        positions: List[List[float]],
        momentums: List[List[float]],
        modes: Optional[Tuple[int, ...]] = None,
    ) -> np.ndarray:
        r"""
        This method calculates the Wigner function values at the specified position and
        momentum vectors, according to the following equation:

        .. math::
            W(r) = \frac{1}{\pi^d \sqrt{\mathrm{det} \sigma}}
                \exp \big (
                    - (r - \mu)^T
                    \sigma^{-1}
                    (r - \mu)
                \big ).

        Args:
            positions (list[list[float]]): List of position vectors.
            momentums (list[list[float]]): List of momentum vectors.
            modes (tuple[int], optional):
                Modes where Wigner function should be calculcated.

        Returns:
            numpy.ndarray:
                The Wigner function values in the shape of a grid specified by the
                input.
        """

        if modes:
            reduced_state = self.reduced(modes)
            return gaussian_wigner_function(
                positions,
                momentums,
                d=reduced_state.d,
                mean=reduced_state.xpxp_mean_vector,
                cov=reduced_state.xpxp_covariance_matrix,
            )

        return gaussian_wigner_function(
            positions,
            momentums,
            d=self.d,
            mean=self.xpxp_mean_vector,
            cov=self.xpxp_covariance_matrix,
        )

    def plot_wigner(
        self,
        positions: List[List[float]],
        momentums: List[List[float]],
        mode: Optional[int] = None,
    ) -> None:
        r"""
        Plots the Wigner function in phase space for a single mode
        using the given positions and momentums.

        Args:
            positions (List[List[float]]): List of list of position values (x-axis).
            momentums (List[List[float]]): List of list of momentum values (p-axis) .
            mode (int, optional):
                Mode where Wigner function should be calculcated.

        Note:
            Only a single mode is supported.

        Returns:
            None: This method generates a plot and does not return a value.

        """
        if self._config.validate and mode is not None and mode > self.d - 1:
            raise InvalidModes(
                f"Mode {mode} is out of range for the state with {self.d} modes."
            )

        if mode is not None:
            mode_tuple = (mode,)
        else:
            mode_tuple = None

        if self._config.validate and self.d != 1 and mode_tuple is None:
            raise InvalidModes(
                "The Wigner function can only be plotted for a single mode: "
                f"modes={mode_tuple} was specified."
            )

        gaussian_wigner_function_values = self.wigner_function(
            positions=positions, momentums=momentums, modes=mode_tuple
        )
        x, p = np.meshgrid(positions, momentums)
        positions = x.tolist()
        momentums = p.tolist()

        plot_wigner_function(
            gaussian_wigner_function_values,
            positions=positions,
            momentums=momentums,
        )

    def get_particle_detection_probability(
        self, occupation_number: np.ndarray
    ) -> float:
        if self._config.validate and len(occupation_number) != self.d:
            raise PiquassoException(
                f"The specified occupation number should have length '{self.d}': "
                f"occupation_number='{occupation_number}'."
            )

        calculation = self._get_density_matrix_calculation()

        return self._connector.np.real(
            calculation.get_density_matrix_element(
                bra=occupation_number,
                ket=occupation_number,
            )
        )

    def get_threshold_detection_probability(
        self, occupation_number: Union[np.ndarray, Tuple[int, ...]]
    ) -> float:
        """Calculates the probability of threshold particle number detection events.

        The occupation number needs to be specified as consecutive 0s and 1s, where 0
        corresponds to no particle detection and 1 to detecting some (>0) particles.

        Args:
            occupation_number (Union[np.ndarray, Tuple[int, ...]]):
                Array of 0s and 1s corresponding to particle detection events per mode.

        Raises:
            PiquassoException: When an invalid input 'occupation_number' is specified.

        Returns:
            float: The threshold particle number detection probability.
        """

        if self._config.validate and len(occupation_number) != self.d:
            raise PiquassoException(
                f"The specified occupation number should have length '{self.d}': "
                f"occupation_number='{occupation_number}'."
            )

        occupation_number_np = np.array(occupation_number)

        if self._config.validate and not np.array_equal(
            occupation_number_np, occupation_number_np.astype(bool)
        ):
            raise PiquassoException(
                f"The specified occupation numbers must only contain '0' or '1': "
                f"occupation_number='{occupation_number}'."
            )

        hbar = self._config.hbar

        if not self._is_displaced():
            return calculate_click_probability_nondisplaced(
                self.xpxp_covariance_matrix / hbar,
                tuple(occupation_number),
            )

        return calculate_click_probability(
            self.xpxp_covariance_matrix / hbar,
            self.xpxp_mean_vector / np.sqrt(hbar),
            tuple(occupation_number),
        )

    @property
    def fock_probabilities(self) -> np.ndarray:
        calculation = self._get_density_matrix_calculation()

        return calculation.get_particle_number_detection_probabilities(
            get_fock_space_basis(d=self.d, cutoff=self._config.cutoff)
        )

    def get_marginal_fock_probabilities(self, modes: Tuple[int, ...]) -> np.ndarray:
        """Return the particle number probabilities on a given subsystem."""

        return self.reduced(modes).fock_probabilities

    def get_xp_string_moment(self, string: List[int]) -> complex:
        r"""Moment corresponding to a product of quadrature operators (XP-string).

        The indices of `string` correspond to the xxpp-ordering, i.e.,
        :math:`0, \dots, d-1` denotes :math:`\hat{x}_1, \dots \hat{x}_d` and
        :math:`d, \dots, 2d-1` denotes :math:`\hat{p}_0, \dots, \hat{p}_d`.

        We use the following formula:

        .. math::
            \left\langle Y_{\alpha_1} Y_{\alpha_2} \cdots Y_{\alpha_n} \right\rangle
            = \sum_{\substack{\text{all pairings}\\ \text{and singletons}}}
            \Biggl(
                \prod_{\text{single }r}
                \mu_{\alpha_r}
            \Biggr)
            \Biggl(
                \prod_{\text{pairing }(i,j)} V_{\alpha_i, \alpha_j}
            \Biggr),

        where :math:`Y_{\alpha_i}` are quadrature operators in xxpp-ordering,
        :math:`\mu` is the mean vector from :meth:`xxpp_mean_vector`, :math:`\Sigma`
        is the covariance matrix from :meth:`xxpp_covariance_matrix`, and

        .. math::
            V &= \Sigma + \frac{i\hbar}{2} \Omega, \\
            \Omega &= \begin{bmatrix}
                0 & I \\
                -I & 0
            \end{bmatrix},

        where :math:`\Omega` encodes the canonical commutation relations.

        The reasoning for this formula goes as follows: suppose that the characteristic
        function giving the Weyl ordering of the quadrature operators is
        :math:`\chi_W(\mathbf{x}, \mathbf{p})`. Then, writing the Weyl operator in terms
        of :math:`\hat{x}_i` and :math:`\hat{p}_i` and using the
        Baker-Campbell-Hausdorff formula, we get that for normal ordering
        (:math:`\hat{x}_i` always in the front) we need to use
        :math:`\chi_W(\mathbf{x}, \mathbf{p}) e^{i \mathbf{x}\cdot\mathbf{p}}`, which
        induces a shift in the covariance matrix
        :math:`\Sigma_{\text{normal}} := \Sigma + \frac{i}{2}X`, where

        .. math::
            X = \begin{bmatrix}
                0 & I \\
                I & 0
            \end{bmatrix}.

        By differentiating the characteristic function correspondng to the normal order,
        one can get the normal-ordered moments. Hence, inserting
        :math:`\Sigma_{\text{normal}}` as the covariance matrix in Isserlis' formula,
        one would get the value for the normal ordered moment. However, the terms
        form :math:`(\Sigma_{\text{normal}})_{i+d, i}` do not appear
        (which is also equal to :math:`(\Sigma_{\text{normal}})_{i, i+d}` due to
        symmetry). Therefore, we can just encode the canonical commutation relations
        by applying the following transformation:

        .. math::
            (\Sigma_{\text{normal}})_{i+d, i} \mapsto
                = (\Sigma_{\text{normal}})_{i, i+d} - i \hbar.

        This yields a formula which works for normal ordered moments, but also
        incorporates the commutation relations, therefore it must work for any ordering.

        Furthermore, by differentiating the normal-ordered characteristic function,
        we get---as an extension to the centered Isserlis' theorem---the following
        formula:

        .. math::
            \left\langle Y_{\alpha_1} Y_{\alpha_2} \cdots Y_{\alpha_n} \right\rangle
            = \sum_{\substack{\text{all pairings}\\ \text{and singletons}}}
            \Biggl(
                \prod_{\text{single }r}
                    \mu_{\alpha_r}
            \Biggr)
            \Biggl(
                \prod_{\text{pairing }(i,j)}
                    (\Sigma_\text{normal})_{\alpha_i, \alpha_j}
            \Biggr),

        and here, one can just replace :math:`\Sigma_\text{normal}` with :math:`V` by
        the previous argument to obtain the formula for all orderings.

        Sources:
            - `Isserlis's theorem <https://en.wikipedia.org/wiki/Isserlis%27s_theorem>`_
            - `Ordered Expansions in Boson Amplitude Operators <https://journals.aps.org/pr/abstract/10.1103/PhysRev.177.1857>`_
            - Section 4.4 from `Quantum Continuous Variables <https://books.google.hu/books/about/Quantum_Continuous_Variables.html?id=bMItDwAAQBAJ>`_
        """  # noqa: E501

        d = self.d
        hbar = self._config.hbar

        first_order_moments = self.xxpp_mean_vector
        cov_xxpp = self.xxpp_covariance_matrix

        second_order_moments = cov_xxpp / 2 + 0.5j * hbar * xp_symplectic_form(d)

        # TODO: Reimplement this using the loop hafnian!
        def recursive(op_list: List[int]) -> complex:
            if not op_list:
                return 1.0

            if len(op_list) == 1:
                return first_order_moments[op_list[0]]

            first = op_list[0]
            rest = op_list[1:]

            total = first_order_moments[first] * recursive(rest)

            for j, second in enumerate(rest):
                remaining = rest[:j] + rest[j + 1 :]

                pair_val = second_order_moments[first, second]

                total += pair_val * recursive(remaining)

            return total

        return recursive(list(string))

    def is_pure(self) -> bool:
        return bool(np.isclose(self.get_purity(), 1.0))

    def get_purity(self) -> float:
        """Purity of the Gaussian state."""

        np = self._connector.np

        return np.real(2**self.d / np.sqrt(np.linalg.det(self.xxpp_covariance_matrix)))

    def purify(self) -> "GaussianState":
        """
        Returns a purification corresponding to the current state on a double-sized
        system.

        Note, that the purification results in a state where the mean vector on the
        auxiliary system is identical with the mean vector in the original system.

        Source:
            - `Evaluating capacities of Bosonic Gaussian channels <https://arxiv.org/abs/quant-ph/9912067>`_

        Returns:
            GaussianState: The purified Gaussian state.
        """  # noqa: E501
        np = self._connector.np
        fallback_np = self._connector.fallback_np

        hbar = self._config.hbar
        cov = self.xpxp_covariance_matrix / hbar

        d = self.d

        indices = xpxp_to_xxpp_indices(d)

        S, D = williamson(cov[fallback_np.ix_(indices, indices)], self._connector)

        beta_diagonals = np.diag(np.sqrt(np.abs(np.diag(D)[:d] ** 2 - 1)))
        zeros = np.zeros_like(beta_diagonals)

        beta_skeleton = np.block([[zeros, beta_diagonals], [beta_diagonals, zeros]])

        indices = xxpp_to_xpxp_indices(d)

        TS = S[indices, :]

        beta = TS @ beta_skeleton @ TS.T

        purified_cov = np.block(
            [
                [cov, beta],
                [beta, cov],
            ]
        )

        purification = GaussianState(
            d=2 * d, connector=self._connector, config=self._config
        )

        purification.xpxp_covariance_matrix = hbar * purified_cov

        mean = self.xpxp_mean_vector
        purification.xpxp_mean_vector = np.concatenate([mean] * 2)

        return purification



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np

import piquasso as pq


pytestmark = pytest.mark.benchmark(
    group="fock-interferometer",
)


@pytest.fixture
def interferometer():
    return np.array(
        [
            [
                -0.11035524 + 0.43053175j,
                0.16672794 - 0.47819775j,
                0.01831264 - 0.07497556j,
                0.44214383 + 0.23308614j,
                0.49732734 - 0.20707815j,
            ],
            [
                -0.2226116 - 0.39735917j,
                0.30063104 - 0.55866103j,
                0.00329231 - 0.28009116j,
                0.23376629 - 0.22854039j,
                -0.43588261 + 0.12139052j,
            ],
            [
                0.24737463 - 0.24638855j,
                -0.31444021 - 0.28467995j,
                -0.16425914 - 0.33655956j,
                -0.26884137 + 0.68435951j,
                -0.07230938 - 0.10989756j,
            ],
            [
                0.34411568 + 0.00861778j,
                -0.26208576 - 0.04842002j,
                0.64971999 - 0.05882925j,
                0.29164646 + 0.0682574j,
                0.01229276 + 0.54314998j,
            ],
            [
                -0.59541546 - 0.01014536j,
                -0.28295784 + 0.10016806j,
                0.57604147 - 0.13381814j,
                -0.07509227 + 0.08557502j,
                -0.12237335 - 0.42143858j,
            ],
        ]
    )


def piquasso_benchmark(benchmark, pq_purefock_simulator, interferometer):
    @benchmark
    def func():
        with pq.Program() as new_program:
            pq.Q() | pq.StateVector(occupation_numbers=(1, 1, 1, 0, 0))
            pq.Q() | pq.Interferometer(interferometer)

        pq_purefock_simulator.execute(new_program)


def strawberryfields_benchmark(benchmark, d, cutoff, interferometer, sf):
    @benchmark
    def func():
        new_program = sf.Program(d)
        new_engine = sf.Engine(backend="fock", backend_options={"cutoff_dim": cutoff})

        with new_program.context as q:
            sf.ops.Fock(1) | q[0]
            sf.ops.Fock(1) | q[1]
            sf.ops.Fock(1) | q[2]
            sf.ops.Fock(0) | q[3]
            sf.ops.Fock(0) | q[4]

            sf.ops.Interferometer(interferometer) | (q[0], q[1], q[2], q[3], q[4])

        new_engine.run(new_program)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import numba as nb


@nb.njit(cache=True)
def xxpp_to_xpxp_indices(d: int) -> np.ndarray:
    r"""
    Indices for basis changing from the xxpp to the xpxp basis.

    Args:
        d (int): The number of modes.

    Returns:
        numpy.ndarray: The basis changing indices.
    """

    indices = np.empty(2 * d, dtype=nb.int32)

    for i in range(d):
        indices[2 * i] = i
        indices[2 * i + 1] = d + i

    return indices


@nb.njit(cache=True)
def xpxp_to_xxpp_indices(d: int) -> np.ndarray:
    r"""
    Indices for basis changing from the xpxp to the xxpp basis.

    Args:
        d (int): The number of modes.

    Returns:
        numpy.ndarray: The basis changing indices.
    """

    indices = np.empty(2 * d, dtype=nb.int32)

    for i in range(d):
        indices[i] = 2 * i
        indices[d + i] = 2 * i + 1

    return indices



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import piquasso as pq
import tensorflow as tf

from scipy.stats import unitary_group


pytestmark = pytest.mark.benchmark(
    group="tf-general",
)


@pytest.fixture
def alpha():
    return 0.01


@pytest.fixture
def r():
    return 0.01


@pytest.fixture
def xi():
    return 0.3


parameters = [(d, unitary_group.rvs(d)) for d in range(3, 5)]


@pytest.mark.parametrize("d, interferometer", parameters)
def piquasso_benchmark(benchmark, d, interferometer, alpha, r, xi):
    @benchmark
    def func():
        alpha_ = tf.Variable(alpha)

        with pq.Program() as program:
            pq.Q(all) | pq.Vacuum()

            for i in range(d):
                pq.Q(i) | pq.Displacement(r=alpha) | pq.Squeezing(r)

            pq.Q(all) | pq.Interferometer(interferometer)

            for i in range(d):
                pq.Q(i) | pq.Kerr(xi)

        simulator_fock = pq.PureFockSimulator(
            d=d, config=pq.Config(cutoff=d), connector=pq.TensorflowConnector()
        )

        with tf.GradientTape() as tape:
            state = simulator_fock.execute(program).state
            mean_photon_number = state.mean_photon_number()

        tape.gradient(mean_photon_number, [alpha_])


@pytest.mark.parametrize("d, interferometer", parameters)
def strawberryfields_benchmark(benchmark, d, interferometer, alpha, r, xi, sf):
    @benchmark
    def func():
        program = sf.Program(d)

        mapping = {}

        alpha_ = tf.Variable(alpha)
        param = program.params("alpha")
        mapping["alpha"] = alpha_

        engine = sf.Engine(backend="tf", backend_options={"cutoff_dim": d})

        with program.context as q:
            for i in range(d):
                sf.ops.Dgate(param) | q[i]
                sf.ops.Sgate(r) | q[i]

            sf.ops.Interferometer(interferometer) | tuple(q[i] for i in range(d))

            for i in range(d):
                sf.ops.Kgate(xi) | q[i]

        with tf.GradientTape() as tape:
            result = engine.run(program, args=mapping)
            state = result.state
            mean = sum([state.mean_photon(mode)[0] for mode in range(d)])

        tape.gradient(mean, [alpha_])



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import random
import numpy as np

from scipy.stats import unitary_group

import piquasso as pq


@pytest.mark.monkey
@pytest.mark.parametrize("d", [2, 3, 4, 5])
def test_random_gaussianstate(d):
    with pq.Program() as prepare_random_program:
        for i in range(d):
            r = np.random.normal()
            phi = random.uniform(0, 2 * np.pi)

            pq.Q(i) | pq.Squeezing(r=r, phi=phi)

        random_unitary = np.array(unitary_group.rvs(d))

        pq.Q(all) | pq.Interferometer(random_unitary)

    with pq.Program() as program:
        pq.Q(all) | prepare_random_program

    simulator = pq.GaussianSimulator(d=d)
    state = simulator.execute(program).state
    state.validate()



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import patch

import pytest

import piquasso as pq


def test_BaseConnector_cannot_be_instantiated():
    with pytest.raises(Exception):
        pq.api.connector.BaseConnector()


def test_BaseConnector_with_overriding_defaults():
    """
    NOTE: This test basically tests Python itself, but it is left here for us to
    remember that the `BaseConnector` class defaults need to be able to overridden for
    any plugin which might want to use e.g. different permanent or hafnian calculation.
    """

    def plugin_loop_hafnian():
        return 43

    class PluginConnector(pq.api.connector.BaseConnector):
        def __init__(self) -> None:
            super().__init__()

            self.loop_hafnian = plugin_loop_hafnian

    p = patch.multiple(PluginConnector, __abstractmethods__=set())

    p.start()
    plugin_connector = PluginConnector()
    p.stop()

    assert plugin_connector.loop_hafnian is plugin_loop_hafnian

    assert plugin_connector.loop_hafnian() == 43


def test_BaseConnector_repr():

    class PluginConnector(pq.api.connector.BaseConnector):
        def __init__(self) -> None:
            super().__init__()

    p = patch.multiple(PluginConnector, __abstractmethods__=set())

    p.start()
    plugin_connector = PluginConnector()
    p.stop()

    assert repr(plugin_connector) == str(plugin_connector) == "PluginConnector()"



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import piquasso as pq

import numpy as np

from jax import grad, jit


def test_mean_photon_number_gradient():
    def func(r):
        simulator = pq.GaussianSimulator(d=2, connector=pq.JaxConnector())

        with pq.Program() as program:
            pq.Q() | pq.Vacuum()

            pq.Q(0) | pq.Squeezing(r)

            pq.Q(0, 1) | pq.Beamsplitter()

        result = simulator.execute(program)

        mean_photon_number = result.state.mean_photon_number(modes=(0, 1))

        return mean_photon_number

    grad_func = jit(grad(func))

    r = 0.3

    assert np.allclose(func(r), np.sinh(r) ** 2)
    assert np.allclose(grad_func(r), 2 * np.sinh(r) * np.cosh(r))


def test_variance_photon_number_gradient():
    def func(r):
        simulator = pq.GaussianSimulator(d=2, connector=pq.JaxConnector())

        with pq.Program() as program:
            pq.Q() | pq.Vacuum()

            pq.Q(0) | pq.Displacement(r)

            pq.Q(0, 1) | pq.Beamsplitter()

        result = simulator.execute(program)

        variance_photon_number = result.state.variance_photon_number(modes=(0, 1))

        return variance_photon_number

    grad_func = jit(grad(func))

    r = 0.3
    assert np.allclose(func(r), r**2)
    assert np.allclose(grad_func(r), 2 * r)


def test_get_particle_number_detection_probability():
    def func(r):
        simulator = pq.GaussianSimulator(d=2, connector=pq.JaxConnector())

        with pq.Program() as program:
            pq.Q() | pq.Vacuum()

            pq.Q(0) | pq.Squeezing(r)

            pq.Q(0, 1) | pq.Beamsplitter()

        result = simulator.execute(program)

        probability = (
            result.state.get_particle_detection_probability((2, 0))
            + result.state.get_particle_detection_probability((1, 1))
            + result.state.get_particle_detection_probability((0, 2))
        )

        return probability

    grad_func = grad(func)

    r = 0.3
    assert np.allclose(func(r), np.tanh(r) ** 2 / (2 * np.cosh(r)))
    assert np.allclose(
        grad_func(r), np.tanh(r) / np.cosh(r) ** 3 - np.tanh(r) ** 3 / np.cosh(r) / 2.0
    )


def test_displaced_get_particle_number_detection_probability_gradient():
    def func(r):
        simulator = pq.GaussianSimulator(d=2, connector=pq.JaxConnector())

        with pq.Program() as program:
            pq.Q() | pq.Vacuum()

            pq.Q(0) | pq.Displacement(r)

            pq.Q(0, 1) | pq.Beamsplitter()

        result = simulator.execute(program)

        probability = (
            result.state.get_particle_detection_probability((2, 0))
            + result.state.get_particle_detection_probability((1, 1))
            + result.state.get_particle_detection_probability((0, 2))
        )

        return probability

    grad_func = grad(func)

    r = 0.3
    assert np.allclose(func(r), r**4 * np.exp(-(r**2)) / 2.0)
    assert np.allclose(grad_func(r), (2 * r**3 - r**5) * np.exp(-(r**2)))



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import piquasso as pq
from piquasso.api.exceptions import InvalidModes


def test_try_registering_nondistinct_modes():
    nondistinct_modes = (0, 0, 1)

    with pq.Program():
        with pytest.raises(InvalidModes):
            pq.Q(*nondistinct_modes) | pq.Fourier()



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import scipy

import numpy as np

from piquasso._math.permanent import permanent as permanent_with_reduction
from piquasso._math.hafnian import (
    hafnian_with_reduction,
    loop_hafnian_with_reduction,
    loop_hafnian_with_reduction_batch,
)

from ..connector import BuiltinConnector, instancemethod

from .interferometer import calculate_interferometer_on_fock_space
from .connections import calculate_interferometer_on_fermionic_fock_space


class NumpyConnector(BuiltinConnector):
    """The calculations for a simulation using NumPy (and SciPy).

    This is enabled by default in the built-in simulators.
    """

    np = fallback_np = forward_pass_np = np

    block_diag = instancemethod(scipy.linalg.block_diag)
    block = instancemethod(np.block)
    logm = instancemethod(scipy.linalg.logm)
    expm = instancemethod(scipy.linalg.expm)
    powm = instancemethod(np.linalg.matrix_power)
    polar = instancemethod(scipy.linalg.polar)
    svd = instancemethod(np.linalg.svd)
    schur = instancemethod(scipy.linalg.schur)
    permanent = instancemethod(permanent_with_reduction)
    hafnian = instancemethod(hafnian_with_reduction)
    loop_hafnian = instancemethod(loop_hafnian_with_reduction)
    loop_hafnian_batch = instancemethod(loop_hafnian_with_reduction_batch)
    calculate_interferometer_on_fock_space = instancemethod(
        calculate_interferometer_on_fock_space
    )
    calculate_interferometer_on_fermionic_fock_space = instancemethod(
        calculate_interferometer_on_fermionic_fock_space
    )

    def sqrtm(self, matrix):
        return scipy.linalg.sqrtm(matrix).astype(np.complex128)

    def preprocess_input_for_custom_gradient(self, value):
        return value

    def assign(self, array, index, value):
        array[index] = value

        return array

    def scatter(self, indices, updates, shape):
        embedded_matrix = np.zeros(shape, dtype=complex)
        indices_array = np.array(indices)
        composite_index = tuple([indices_array[:, i] for i in range(len(shape))])

        embedded_matrix[composite_index] = np.array(updates)

        return embedded_matrix



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

import piquasso as pq


def test_Attenuator_with_zero_theta_changes_nothing_on_one_mode_state():
    with pq.Program() as empty_program:
        pq.Q() | pq.StateVector([1]) / np.sqrt(2)
        pq.Q() | pq.StateVector([2]) / np.sqrt(2)

    with pq.Program() as program_with_zero_theta:
        pq.Q() | pq.StateVector([1]) / np.sqrt(2)
        pq.Q() | pq.StateVector([2]) / np.sqrt(2)

        pq.Q() | pq.Attenuator(theta=0.0)

    simulator = pq.PureFockSimulator(d=1, config=pq.Config(cutoff=3))

    empty_state = simulator.execute(empty_program).state
    lossy_state = simulator.execute(program_with_zero_theta).state

    assert np.allclose(
        empty_state.fock_probabilities,
        lossy_state.fock_probabilities,
    )


def test_Attenuator_with_pi_over_2_theta_maps_to_vacuum():
    with pq.Program() as program_with_zero_theta:
        pq.Q() | pq.StateVector([1]) / np.sqrt(2)
        pq.Q() | pq.StateVector([2]) / np.sqrt(2)

        pq.Q() | pq.Attenuator(theta=np.pi / 2)

    simulator = pq.PureFockSimulator(d=1, config=pq.Config(cutoff=3))

    lossy_state = simulator.execute(program_with_zero_theta).state

    assert np.allclose(
        lossy_state.fock_probabilities,
        [1.0, 0.0, 0.0],
        atol=1e-7,
    )


def test_Attenuator_for_one_particle():
    theta = np.pi / 5

    transmittivity = np.cos(theta)

    with pq.Program() as program:
        pq.Q(0, 1) | pq.StateVector([0, 1])

        pq.Q(1) | pq.Attenuator(theta=theta)

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=2))

    state = simulator.execute(program).state

    assert np.allclose(
        state.fock_probabilities,
        [1 - transmittivity**2, 0, transmittivity**2],
    )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import numpy as np
from scipy.linalg import schur
from piquasso._math.pfaffian import pfaffian


def pfaffian_naive(matrix):
    if matrix.shape[0] == 0:
        return 1.0

    if matrix.shape[0] % 2 == 1:
        return 0.0

    blocks, O = schur(matrix)
    a = np.diag(blocks, 1)[::2]

    return np.prod(a) * np.linalg.det(O)


def test_pfaffian_empty():
    matrix = np.array([[]], dtype=float)

    assert np.isclose(pfaffian(matrix), 1.0)


def test_pfaffian_2_by_2_skew_symmetric_float32():
    matrix = np.array(
        [
            [0, 1.7],
            [-1.7, 0],
        ],
        dtype=np.float32,
    )

    result = pfaffian_naive(matrix)
    output = pfaffian(matrix)

    assert output.dtype == np.float32

    assert np.isclose(output, result)
    assert np.isclose(output, 1.7)


def test_pfaffian_2_by_2_skew_symmetric_float64():
    matrix = np.array(
        [
            [0, 1.7],
            [-1.7, 0],
        ],
        dtype=np.float64,
    )

    output = pfaffian(matrix)

    assert output.dtype == np.float64

    assert np.isclose(output.item(), pfaffian_naive(matrix))
    assert np.isclose(output.item(), 1.7)


@pytest.mark.monkey
def test_pfaffian_4_by_4_skew_symmetric_random():
    for _ in range(100):

        A = np.random.rand(4, 4)
        matrix = A - A.T

        result = pfaffian_naive(matrix)
        output = pfaffian(matrix)

        assert np.isclose(output, result)


@pytest.mark.monkey
def test_pfaffian_6_by_6_skew_symmetric_random():
    for _ in range(100):

        A = np.random.rand(6, 6)
        matrix = A - A.T

        result = pfaffian_naive(matrix)
        output = pfaffian(matrix)

        assert np.isclose(output, result)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np

import piquasso as pq

from pytest_lazy_fixtures import lf


tf_purefock_simulators = (
    lf("PureFockSimulator_with_tensorflow"),
    lf("PureFockSimulator_with_tensorflow_tf_function"),
)

jax_purefock_simulator = [
    lf("PureFockSimulator_with_jax"),
]


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.PureFockSimulator,
        *tf_purefock_simulators,
        *jax_purefock_simulator,
        pq.FockSimulator,
    ),
)
def test_squeezing_probabilities(SimulatorClass):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Squeezing(r=0.5, phi=np.pi / 3)

    simulator = SimulatorClass(d=2, config=pq.Config(cutoff=3))
    state = simulator.execute(program).state

    state.normalize()

    state.validate()

    assert np.isclose(
        sum(state.fock_probabilities), 1.0
    ), "The state should be renormalized for probability conservation."

    assert np.allclose(
        state.fock_probabilities, [0.90352508, 0.0, 0.0, 0.09647492, 0.0, 0.0]
    )


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.PureFockSimulator,
        *tf_purefock_simulators,
        *jax_purefock_simulator,
        pq.FockSimulator,
    ),
)
def test_displacement_probabilities(SimulatorClass):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=0.5, phi=np.pi / 3)

    simulator = SimulatorClass(d=2, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    state.normalize()

    state.validate()

    assert np.isclose(
        sum(state.fock_probabilities), 1.0
    ), "The state should be renormalized for probability conservation."

    assert np.allclose(
        state.fock_probabilities, [0.7804878, 0.19512195, 0.0, 0.02439024, 0.0, 0.0]
    )


@pytest.mark.parametrize("SimulatorClass", (pq.PureFockSimulator, pq.FockSimulator))
def test_displacement_get_fock_prob(SimulatorClass):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=0.2, phi=np.pi / 3)

    simulator = SimulatorClass(d=2, config=pq.Config(cutoff=8))

    state = simulator.execute(program).state

    state.normalize()

    state.validate()
    assert np.isclose(
        state.get_particle_detection_probability(occupation_number=(0, 1)), 0.0
    )


def test_PureFockState_displacement_on_single_mode():
    alpha = 2.0
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=alpha)

    simulator = pq.PureFockSimulator(d=1, config=pq.Config(cutoff=20))

    state = simulator.execute(program).state

    # TODO: Better way of presenting the resulting state.
    nonzero_elements = list(state.nonzero_elements)

    assert len(nonzero_elements) == 20
    np.isclose(nonzero_elements[0][0], np.exp(-0.5 * np.abs(alpha**2)))


def test_FockState_displacement_on_single_mode():
    alpha = 2.0
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=alpha)

    simulator = pq.FockSimulator(d=1, config=pq.Config(cutoff=20))

    state = simulator.execute(program).state

    # TODO: Better way of presenting the resulting state.
    nonzero_elements = list(state.nonzero_elements)

    assert len(nonzero_elements) == 400.0
    np.isclose(nonzero_elements[0][0], np.exp(-0.5 * np.abs(alpha**2)))


def test_PureFockState_squeezing_on_single_mode():
    r = 0.5
    phi = np.pi / 3

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Squeezing(r=r, phi=phi)

    simulator = pq.PureFockSimulator(d=1, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    # TODO: Better way of presenting the resulting state.
    nonzero_elements = list(state.nonzero_elements)

    assert len(nonzero_elements) == 2

    assert nonzero_elements[0][1] == (0,)
    assert np.isclose(nonzero_elements[0][0], 1 / np.sqrt(np.cosh(r)))

    assert nonzero_elements[1][1] == (2,)
    assert np.isclose(
        nonzero_elements[1][0],
        (-np.exp(1j * phi) * np.tanh(r) / np.sqrt(2)) / np.sqrt(np.cosh(r)),
    )


def test_PureFockState_squeezing():
    r = 0.5
    phi = np.pi / 3

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Squeezing(r=r, phi=phi)

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    # TODO: Better way of presenting the resulting state.
    nonzero_elements = list(state.nonzero_elements)

    assert len(nonzero_elements) == 2

    assert nonzero_elements[0][1] == (0, 0)
    assert np.isclose(nonzero_elements[0][0], 1 / np.sqrt(np.cosh(r)))

    assert nonzero_elements[1][1] == (2, 0)
    assert np.isclose(
        nonzero_elements[1][0],
        (-np.exp(1j * phi) * np.tanh(r) / np.sqrt(2)) / np.sqrt(np.cosh(r)),
    )


def test_PureFockState_displacement():
    alpha = 0.5 * np.exp(1j * np.pi / 3)

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(0.5, np.pi / 3)

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    # TODO: Better way of presenting the resulting state.
    nonzero_elements = list(state.nonzero_elements)

    assert len(nonzero_elements) == 3

    assert nonzero_elements[0][1] == (0, 0)
    assert np.isclose(nonzero_elements[0][0], np.exp(-np.abs(alpha) ** 2 / 2))

    assert nonzero_elements[1][1] == (1, 0)
    assert np.isclose(nonzero_elements[1][0], np.exp(-np.abs(alpha) ** 2 / 2) * alpha)

    assert nonzero_elements[2][1] == (2, 0)
    assert np.isclose(
        nonzero_elements[2][0],
        np.exp(-np.abs(alpha) ** 2 / 2) * (alpha**2) / np.sqrt(2),
    )


def test_FockState_squeezing():
    r = 0.5
    phi = np.pi / 3

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Squeezing(r=r, phi=phi)

    simulator = pq.FockSimulator(d=2, config=pq.Config(cutoff=3))
    state = simulator.execute(program).state

    # TODO: Better way of presenting the resulting state.
    nonzero_elements = list(state.nonzero_elements)

    vacuum_probability = 1 / np.cosh(r)

    two_particle_probability = (1 / np.cosh(r)) * np.tanh(r) ** 2 / 2

    assert len(nonzero_elements) == 4

    assert nonzero_elements[0][1] == ((0, 0), (0, 0))
    assert np.isclose(nonzero_elements[0][0], vacuum_probability)

    assert nonzero_elements[1][1] == ((0, 0), (2, 0))
    assert np.isclose(
        nonzero_elements[1][0],
        (-np.exp(-1j * phi) * np.tanh(r) / np.sqrt(2)) / np.cosh(r),
    )

    assert nonzero_elements[2][1] == ((2, 0), (0, 0))
    assert np.isclose(
        nonzero_elements[2][0],
        (-np.exp(1j * phi) * np.tanh(r) / np.sqrt(2)) / np.cosh(r),
    )

    assert nonzero_elements[3][1] == ((2, 0), (2, 0))
    assert np.isclose(nonzero_elements[3][0], two_particle_probability)


def test_FockState_displacement():
    alpha = 0.5 * np.exp(1j * np.pi / 3)

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(0.5, np.pi / 3)

    simulator = pq.FockSimulator(d=2, config=pq.Config(cutoff=2))
    state = simulator.execute(program).state

    # TODO: Better way of presenting the resulting state.
    nonzero_elements = list(state.nonzero_elements)

    assert len(nonzero_elements) == 4

    vacuum_probability = np.exp(-np.abs(alpha) ** 2)

    one_particle_probability = np.exp(-np.abs(alpha) ** 2) * np.abs(alpha) ** 2

    assert nonzero_elements[0][1] == ((0, 0), (0, 0))
    assert np.isclose(nonzero_elements[0][0], vacuum_probability)

    assert nonzero_elements[1][1] == ((0, 0), (1, 0))
    assert np.isclose(
        nonzero_elements[1][0],
        np.exp(-np.abs(alpha) ** 2) * alpha.conj(),
    )

    assert nonzero_elements[2][1] == ((1, 0), (0, 0))
    assert np.isclose(nonzero_elements[2][0], np.exp(-np.abs(alpha) ** 2) * alpha)

    assert nonzero_elements[3][1] == ((1, 0), (1, 0))
    assert np.isclose(
        nonzero_elements[3][0],
        one_particle_probability,
    )


def test_FockState_cubic_phase():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=0.2, phi=0)
        pq.Q(0) | pq.CubicPhase(gamma=1.0)

    simulator = pq.FockSimulator(d=1, config=pq.Config(cutoff=5))
    state = simulator.execute(program).state
    state.normalize()

    x_mean, x_var = state.quadratures_mean_variance(modes=(0,))
    p_mean, p_var = state.quadratures_mean_variance(modes=(0,), phi=np.pi / 2)

    assert np.isclose(x_mean, 0.4, atol=1e-4)
    assert np.isclose(x_var, 1.0, atol=1e-4)
    assert np.isclose(p_mean, 0.51863789, atol=1e-4)
    assert np.isclose(p_var, 2.08869536, atol=1e-4)


def test_PureFockState_cubic_phase():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=0.2, phi=0)
        pq.Q(1) | pq.CubicPhase(gamma=1.0)

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=5))
    state = simulator.execute(program).state
    state.normalize()

    x_mean, x_var = state.quadratures_mean_variance(modes=(0,))
    p_mean, p_var = state.quadratures_mean_variance(modes=(1,), phi=np.pi / 2)

    assert np.isclose(x_mean, 0.3929401, atol=1e-4)
    assert np.isclose(x_var, 0.999321, atol=1e-4)
    assert np.isclose(p_mean, 0.547729, atol=1e-4)
    assert np.isclose(p_var, 1.8674705, atol=1e-4)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

import pytest

import numpy as np

from unittest import mock


def test_import_piquasso_works_without_tensorflow_dependency():
    with mock.patch.dict(sys.modules, {"tensorflow": None}):
        import piquasso

        print(piquasso)


def test_import_piquasso_works_without_jax_dependency():
    with mock.patch.dict(sys.modules, {"jax": None}):
        import piquasso

        print(piquasso)


def test_matplotlib_plot_importerror():
    with mock.patch.dict("sys.modules", {"matplotlib.pyplot": None}):
        from piquasso._simulators.plot import plot_wigner_function
        import pytest

        with pytest.raises(
            ImportError, match="The visualization feature requires matplotlib."
        ):
            plot_wigner_function(
                vals=[[0, 1], [2, 3]],
                positions=[[0, 1], [2, 3]],
                momentums=[[0, 1], [2, 3]],
            )


def test_use_dask_without_dask_raises_ImportError():
    with mock.patch.dict("sys.modules", {"dask": None}):
        import piquasso as pq

        A = np.array(
            [
                [0, 1, 0, 1, 1],
                [1, 0, 0, 0, 1],
                [0, 0, 0, 1, 0],
                [1, 0, 1, 0, 1],
                [1, 1, 0, 1, 0],
            ]
        )

        with pq.Program() as gaussian_boson_sampling:
            pq.Q(all) | pq.Graph(A)

            pq.Q(all) | pq.ParticleNumberMeasurement()

        simulator = pq.GaussianSimulator(
            d=len(A),
            config=pq.Config(use_dask=True),
        )

        with pytest.raises(
            ImportError, match="This feature requires 'dask' to be installed."
        ):
            simulator.execute(gaussian_boson_sampling, shots=50)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, TYPE_CHECKING

from .._utils import (
    get_cutoff_fock_space_dimension,
    get_fock_space_index,
    get_fock_space_basis,
)

from piquasso.api.exceptions import InvalidState

from piquasso.api.state import State
from piquasso.api.config import Config
from piquasso.api.connector import BaseConnector

if TYPE_CHECKING:
    import numpy as np


class PureFockState(State):
    r"""A fermionic pure Fock state."""

    def __init__(
        self,
        d: int,
        connector: BaseConnector,
        config: Optional[Config] = None,
    ) -> None:
        self._d = d

        super().__init__(connector=connector, config=config)

        dim = get_cutoff_fock_space_dimension(d, self._config.cutoff)

        self._state_vector = connector.np.zeros(
            shape=dim, dtype=self._config.complex_dtype
        )

    @property
    def state_vector(self):
        r"""The state vector of the quantum state.

        .. warning::
            The primary ordering of the Fock basis is by number of particles, and the
            secondary is anti-lexicographic.

            Example for 3 modes:

            .. math ::

                \ket{000},
                \ket{100}, \ket{010}, \ket{001},
                \ket{200}, \ket{110}, \ket{101}, \ket{020}, \ket{011}, \ket{002}, \dots

            This is in contrast with
            :meth:`~piquasso.fermionic.gaussian.state.GaussianState.density_matrix`,
            where the primary ordering is lexicographic.
        """
        return self._state_vector

    @property
    def d(self):
        return self._d

    @property
    def fock_probabilities(self) -> "np.ndarray":
        np = self._connector.np
        return np.conj(self._state_vector) * self._state_vector

    @property
    def fock_probabilities_map(self) -> dict:
        fock_space_basis = get_fock_space_basis(self.d, self._config.cutoff)

        occupation_numbers = [tuple(x) for x in fock_space_basis.tolist()]

        return dict(zip(occupation_numbers, self.fock_probabilities))

    @property
    def norm(self) -> float:
        return self._connector.np.sum(self.fock_probabilities)

    def get_particle_detection_probability(self, occupation_number):
        index = get_fock_space_index(
            self._connector.fallback_np.array(occupation_number, dtype=int)
        )

        return self._connector.np.abs(self._state_vector[index]) ** 2

    def validate(self):
        if not self._config.validate:
            return

        np = self._connector.np

        norm = self.norm

        if not np.isclose(norm, 1.0):
            raise InvalidState(
                f"The sum of probabilities is {norm}, instead of 1.0:\n"
                f"fock_probabilities={self.fock_probabilities}"
            )

    @property
    def density_matrix(self):
        return self._connector.np.outer(self._state_vector, self._state_vector.conj())

    def __eq__(self, other):
        if not isinstance(other, PureFockState):
            return False

        return self._np.allclose(self.state_vector, other.state_vector)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import piquasso as pq


def test_GaussianSimulator_supports_NumpyConnector():
    pq.GaussianSimulator(d=1, connector=pq.NumpyConnector())


def test_GaussianSimulator_does_not_support_TensorflowConnector(tf):
    connector = pq.TensorflowConnector()

    with pytest.raises(pq.api.exceptions.InvalidSimulation) as error:
        pq.GaussianSimulator(d=1, connector=connector)

    assert f"The connector '{connector}' is not supported." in error.value.args[0]


def test_SamplingSimulator_supports_NumpyConnector():
    pq.SamplingSimulator(d=1, connector=pq.NumpyConnector())


def test_SamplingSimulator_does_not_support_TensorflowConnector(tf):
    connector = pq.TensorflowConnector()

    with pytest.raises(pq.api.exceptions.InvalidSimulation) as error:
        pq.SamplingSimulator(d=1, connector=connector)

    assert f"The connector '{connector}' is not supported." in error.value.args[0]


def test_FockSimulator_supports_NumpyConnector():
    pq.FockSimulator(d=1, connector=pq.NumpyConnector())


def test_FockSimulator_does_not_support_TensorflowConnector(tf):
    connector = pq.TensorflowConnector()

    with pytest.raises(pq.api.exceptions.InvalidSimulation) as error:
        pq.FockSimulator(d=1, connector=connector)

    assert f"The connector '{connector}' is not supported." in error.value.args[0]


def test_PureFockSimulator_supports_NumpyConnector():
    pq.PureFockSimulator(d=1, connector=pq.NumpyConnector())


def test_PureFockSimulator_supports_TensorflowConnector(tf):
    pq.PureFockSimulator(d=1, connector=pq.TensorflowConnector())



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple

import numpy as np
import numba as nb

from piquasso._math.combinatorics import arr_comb, comb


def get_operator_index(modes: Tuple[int, ...]) -> Tuple[np.ndarray, np.ndarray]:
    """
    Note:
        For indexing of numpy arrays, see
        https://numpy.org/doc/stable/reference/arrays.indexing.html#advanced-indexing
    """

    transformed_columns = np.array([modes] * len(modes))
    transformed_rows = transformed_columns.transpose()

    return transformed_rows, transformed_columns


def get_auxiliary_operator_index(
    modes: Tuple[int, ...], auxiliary_modes: Tuple[int, ...]
) -> Tuple[Tuple[int, ...], Tuple[int, ...]]:
    auxiliary_rows = tuple(np.array([modes] * len(auxiliary_modes)).transpose())

    return auxiliary_rows, auxiliary_modes


@nb.njit(cache=True)
def get_index_in_fock_space(element):
    sum_ = 0
    accumulator = 0
    for i in range(len(element)):
        sum_ += element[-1 - i]
        accumulator += comb(sum_ + i, i + 1)

    return accumulator


@nb.njit(cache=True)
def get_index_in_fock_space_array(basis: np.ndarray) -> np.ndarray:
    sum_ = np.zeros(shape=basis.shape[:-1], dtype=np.int32)
    accumulator = np.zeros(shape=basis.shape[:-1], dtype=np.int32)

    for i in range(basis.shape[-1]):
        sum_ += basis[..., -1 - i]
        accumulator += arr_comb(sum_ + i, i + 1)

    return accumulator


@nb.njit(cache=True)
def get_index_in_fock_subspace(element: np.ndarray) -> int:
    sum_ = 0
    accumulator = 0
    for i in range(len(element) - 1):
        sum_ += element[-1 - i]
        accumulator += comb(sum_ + i, i + 1)

    return accumulator


@nb.njit(cache=True)
def get_index_in_fock_subspace_array(basis: np.ndarray) -> np.ndarray:
    sum_ = np.zeros(shape=basis.shape[:-1], dtype=np.int32)
    accumulator = np.zeros(shape=basis.shape[:-1], dtype=np.int32)

    for i in range(basis.shape[-1] - 1):
        sum_ += basis[..., -1 - i]
        accumulator += arr_comb(sum_ + i, i + 1)

    return accumulator


@nb.njit(cache=True)
def get_auxiliary_modes(d: int, modes: Tuple[int, ...]) -> np.ndarray:
    return np.delete(np.arange(d), modes)


@nb.njit(cache=True)
def double_modes(arr):
    result = np.empty(len(arr) * 2, dtype=arr.dtype)
    result[0::2] = 2 * arr
    result[1::2] = 2 * arr + 1
    return result



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The Piquasso module.

One can access all the instructions and states from here as attributes.
"""

from piquasso import cvqnn
from piquasso import fermionic

from piquasso.api.mode import Q
from piquasso.api.config import Config
from piquasso.api.instruction import (
    Instruction,
    Preparation,
    Gate,
    Measurement,
)
from piquasso.api.program import Program
from piquasso.api.state import State
from piquasso.api.computer import Computer
from piquasso.api.simulator import Simulator
from piquasso.api.utils import as_code

from piquasso._simulators.sampling import SamplingState, SamplingSimulator

from piquasso._simulators.gaussian import GaussianState, GaussianSimulator
from piquasso._simulators.fock import (
    FockState,
    PureFockState,
    BatchPureFockState,
    FockSimulator,
    PureFockSimulator,
)

from piquasso._simulators.connectors import (
    NumpyConnector,
    TensorflowConnector,
    JaxConnector,
)

from .instructions.preparations import (
    Vacuum,
    Mean,
    Covariance,
    Thermal,
    StateVector,
    DensityMatrix,
    Create,
    Annihilate,
)

from .instructions.gates import (
    GaussianTransform,
    Phaseshifter,
    Beamsplitter,
    Beamsplitter5050,
    MachZehnder,
    Fourier,
    Displacement,
    PositionDisplacement,
    MomentumDisplacement,
    Squeezing,
    QuadraticPhase,
    Squeezing2,
    Kerr,
    CrossKerr,
    ControlledX,
    ControlledZ,
    Interferometer,
    Graph,
    CubicPhase,
)

from .instructions.measurements import (
    ParticleNumberMeasurement,
    ThresholdMeasurement,
    HomodyneMeasurement,
    HeterodyneMeasurement,
    GeneraldyneMeasurement,
    PostSelectPhotons,
    ImperfectPostSelectPhotons,
)

from .instructions.channels import (
    DeterministicGaussianChannel,
    Attenuator,
    Loss,
    LossyInterferometer,
)

from .instructions.batch import (
    BatchPrepare,
    BatchApply,
)


__all__ = [
    # API
    "Program",
    "Q",
    "Config",
    "Instruction",
    "Preparation",
    "Gate",
    "Measurement",
    "State",
    "Computer",
    "Simulator",
    "as_code",
    # Simulators
    "GaussianSimulator",
    "SamplingSimulator",
    "FockSimulator",
    "PureFockSimulator",
    # Connectors
    "NumpyConnector",
    "TensorflowConnector",
    "JaxConnector",
    # States
    "GaussianState",
    "SamplingState",
    "FockState",
    "PureFockState",
    "BatchPureFockState",
    # Preparations
    "Vacuum",
    "Mean",
    "Covariance",
    "Thermal",
    "StateVector",
    "DensityMatrix",
    "Create",
    "Annihilate",
    # Gates
    "GaussianTransform",
    "Phaseshifter",
    "Beamsplitter",
    "Beamsplitter5050",
    "MachZehnder",
    "Fourier",
    "Displacement",
    "PositionDisplacement",
    "MomentumDisplacement",
    "Squeezing",
    "QuadraticPhase",
    "Squeezing2",
    "Kerr",
    "CrossKerr",
    "CubicPhase",
    "ControlledX",
    "ControlledZ",
    "Interferometer",
    "Graph",
    # Measurements
    "ParticleNumberMeasurement",
    "ThresholdMeasurement",
    "HomodyneMeasurement",
    "HeterodyneMeasurement",
    "GeneraldyneMeasurement",
    "PostSelectPhotons",
    "ImperfectPostSelectPhotons",
    # Channels
    "DeterministicGaussianChannel",
    "Attenuator",
    "Loss",
    "LossyInterferometer",
    # Batch
    "BatchPrepare",
    "BatchApply",
    # Modules
    "cvqnn",
    "fermionic",
]

__version__ = "6.2.0"



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from functools import lru_cache

from scipy.special import factorial

from piquasso.api.connector import BaseConnector


def create_single_mode_displacement_matrix(
    r: float,
    phi: float,
    cutoff: int,
    complex_dtype: np.dtype,
    connector: BaseConnector,
) -> np.ndarray:
    r"""
    This method generates the Displacement operator following a recursion rule.
    Reference: https://quantum-journal.org/papers/q-2020-11-30-366/.
    Args:
    r (float): This is the Displacement amplitude. Typically this value can be
        negative or positive depending on the desired displacement direction.
        Note:
            Setting :math:`|r|` to higher values will require you to have a higher
            cutoff dimensions.
    phi (float): This is the Dispalacement angle. Its ranges are
        :math:`\phi \in [ 0, 2 \pi )`
    Returns:
        np.ndarray: The constructed Displacement matrix representing the Fock
        operator.
    """
    np = connector.forward_pass_np
    fallback_np = connector.fallback_np

    cutoff_range = fallback_np.arange(cutoff)
    sqrt_indices = fallback_np.sqrt(cutoff_range)
    denominator = 1 / fallback_np.sqrt(factorial(cutoff_range))

    displacement = r * np.exp(1j * phi)
    displacement_conj = np.conj(displacement)

    matrix = connector.accumulator(dtype=complex_dtype, size=cutoff)

    # NOTE: Tensorflow does not implement the NumPy API correctly, since in
    # tensorflow `np.power(0.0j, 0.0)` results in `nan+nanj`, whereas in NumPy it
    # is just 1. Instead of redefining `power` we just add a small `epsilon`, which
    # magically yields the correct result.
    epsilon = 10e-100
    previous_element = np.power(displacement + epsilon, cutoff_range) * denominator

    matrix = connector.write_to_accumulator(matrix, 0, previous_element)

    roll_index = fallback_np.arange(-1, cutoff - 1)

    for i in connector.range(1, cutoff):
        previous_element = (
            sqrt_indices * previous_element[roll_index]
            - displacement_conj * previous_element
        )

        matrix = connector.write_to_accumulator(matrix, i, previous_element)

    return (
        np.exp(-0.5 * r**2)
        * connector.transpose(connector.stack_accumulator(matrix))
        * denominator
    )


@lru_cache
def _double_factorial_array(cutoff):
    """
    NOTE: For some reason, this rudimentary implementation is faster than using
    `scipy.special.factorial2`.
    """
    array = np.empty(shape=(cutoff + 1) // 2)

    array[0] = 1

    for i in range(1, len(array)):
        array[i] = (2 * i - 1) / (2 * i) * array[i - 1]

    return array


def create_single_mode_squeezing_matrix(
    r: float,
    phi: float,
    cutoff: int,
    complex_dtype: np.dtype,
    connector: BaseConnector,
) -> np.ndarray:
    r"""
    This method generates the Squeezing operator following a recursion rule.
    Reference: https://quantum-journal.org/papers/q-2020-11-30-366/.

    Args:
    r (float): This is the Squeezing amplitude. Typically this value can be
        negative or positive depending on the desired squeezing direction.
        Note:
            Setting :math:`|r|` to higher values will require you to have a higher
            cutoff dimensions.
    phi (float): This is the Squeezing angle. Its ranges are
        :math:`\phi \in [ 0, 2 \pi )`

    Returns:
        np.ndarray: The constructed Squeezing matrix representing the Fock operator.
    """

    np = connector.forward_pass_np
    fallback_np = connector.fallback_np

    sechr = 1.0 / np.cosh(r)
    A = np.exp(1j * phi) * np.tanh(r)
    Aconj = np.conj(A)
    sqrt_indices = np.sqrt(fallback_np.arange(cutoff))
    sechr_sqrt_indices = sechr * sqrt_indices
    A_conj_sqrt_indices = Aconj * sqrt_indices

    # NOTE: Tensorflow does not implement the NumPy API correctly, since in
    # tensorflow `np.power(0.0j, 0.0)` results in `nan+nanj`, whereas in NumPy it
    # is just 1. Instead of redefining `power` we just add a small `epsilon`, which
    # magically yields the correct result.
    epsilon = 10e-100
    first_row_nonzero = fallback_np.sqrt(_double_factorial_array(cutoff)) * np.power(
        -(A + epsilon), fallback_np.arange(0, (cutoff + 1) // 2)
    )

    first_row = np.zeros(shape=cutoff, dtype=complex_dtype)
    first_row = connector.assign(
        first_row, fallback_np.arange(0, cutoff, 2), first_row_nonzero
    )

    roll_index = fallback_np.arange(-1, cutoff - 1)
    second_row = sechr_sqrt_indices * first_row[roll_index]

    matrix = connector.accumulator(dtype=complex_dtype, size=cutoff)

    matrix = connector.write_to_accumulator(matrix, 0, first_row)
    matrix = connector.write_to_accumulator(matrix, 1, second_row)

    previous_previous = first_row
    previous = second_row

    for col in connector.range(2, cutoff):
        current = (
            sechr_sqrt_indices * previous[roll_index]
            + A_conj_sqrt_indices[col - 1] * previous_previous
        ) / sqrt_indices[col]

        matrix = connector.write_to_accumulator(matrix, col, current)
        previous_previous = previous
        previous = current

    return np.sqrt(sechr) * connector.transpose(connector.stack_accumulator(matrix))



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

import piquasso as pq


def test_measure_particle_number_on_one_mode():
    with pq.Program() as program:
        pq.Q() | pq.StateVector([0, 1, 1]) * np.sqrt(2 / 6)

        pq.Q(2) | pq.StateVector([1]) * np.sqrt(1 / 6)
        pq.Q(2) | pq.StateVector([2]) * np.sqrt(3 / 6)

        pq.Q(2) | pq.ParticleNumberMeasurement()

    simulator = pq.PureFockSimulator(d=3)

    result = simulator.execute(program)

    assert np.isclose(sum(result.state.fock_probabilities), 1)

    sample = result.samples[0]
    assert sample == (1,) or sample == (2,)

    if sample == (1,):
        expected_simulator = pq.PureFockSimulator(d=3)
        expected_state = expected_simulator.execute_instructions(
            instructions=[
                0.5773502691896258 * pq.StateVector([0, 0, 1]),
                0.816496580927726 * pq.StateVector([0, 1, 1]),
            ]
        ).state

    elif sample == (2,):
        expected_simulator = pq.PureFockSimulator(d=3)
        expected_state = expected_simulator.execute_instructions(
            instructions=[pq.StateVector([0, 0, 2])]
        ).state

    assert result.state == expected_state


def test_measure_particle_number_on_two_modes():
    with pq.Program() as program:
        pq.Q(1, 2) | pq.StateVector([1, 1]) * np.sqrt(2 / 6)
        pq.Q(1, 2) | pq.StateVector([0, 1]) * np.sqrt(1 / 6)
        pq.Q(1, 2) | pq.StateVector([0, 2]) * np.sqrt(3 / 6)

        pq.Q(1, 2) | pq.ParticleNumberMeasurement()

    simulator = pq.PureFockSimulator(d=3)

    result = simulator.execute(program)

    assert np.isclose(sum(result.state.fock_probabilities), 1)

    sample = result.samples[0]
    assert sample == (0, 1) or sample == (1, 1) or sample == (0, 2)

    if sample == (0, 1):
        expected_simulator = pq.PureFockSimulator(d=3)
        expected_state = expected_simulator.execute_instructions(
            instructions=[pq.StateVector([0, 0, 1])]
        ).state

    elif sample == (1, 1):
        expected_simulator = pq.PureFockSimulator(d=3)
        expected_state = expected_simulator.execute_instructions(
            instructions=[pq.StateVector([0, 1, 1])]
        ).state

    elif sample == (0, 2):
        expected_simulator = pq.PureFockSimulator(d=3)
        expected_state = expected_simulator.execute_instructions(
            instructions=[pq.StateVector([0, 0, 2])]
        ).state

    assert result.state == expected_state


def test_measure_particle_number_on_all_modes():
    config = pq.Config(cutoff=2)

    simulator = pq.PureFockSimulator(d=3, config=config)

    with pq.Program() as program:
        pq.Q() | 0.5 * pq.StateVector([0, 0, 0])
        pq.Q() | 0.5 * pq.StateVector([0, 0, 1])
        pq.Q() | np.sqrt(1 / 2) * pq.StateVector([1, 0, 0])

        pq.Q() | pq.ParticleNumberMeasurement()

    result = simulator.execute(program)

    assert np.isclose(sum(result.state.fock_probabilities), 1)

    sample = result.samples[0]
    assert sample == (0, 0, 0) or sample == (1, 0, 0) or sample == (0, 0, 1)

    if sample == (0, 0, 0):
        expected_simulator = pq.PureFockSimulator(d=3, config=config)
        expected_state = expected_simulator.execute_instructions(
            instructions=[
                pq.StateVector([0, 0, 0]),
            ],
        ).state

    elif sample == (0, 0, 1):
        expected_simulator = pq.PureFockSimulator(d=3, config=config)
        expected_state = expected_simulator.execute_instructions(
            instructions=[
                pq.StateVector([0, 0, 1]),
            ]
        ).state

    elif sample == (1, 0, 0):
        expected_simulator = pq.PureFockSimulator(d=3, config=config)
        expected_state = expected_simulator.execute_instructions(
            instructions=[
                pq.StateVector([1, 0, 0]),
            ],
        ).state

    assert result.state == expected_state


def test_measure_particle_number_with_multiple_shots():
    shots = 4

    # TODO: This is very unusual, that we need to know the cutoff for specifying the
    # state. It should be imposed, that the only parameter for a state should be `d` and
    #  `config` maybe.
    simulator = pq.PureFockSimulator(d=3, config=pq.Config(cutoff=2))

    with pq.Program() as program:
        pq.Q() | 0.5 * pq.StateVector([0, 0, 0])
        pq.Q() | 0.5 * pq.StateVector([0, 0, 1])
        pq.Q() | np.sqrt(1 / 2) * pq.StateVector([1, 0, 0])

        pq.Q() | pq.ParticleNumberMeasurement()

    result = simulator.execute(program, shots)

    assert np.isclose(sum(result.state.fock_probabilities), 1)
    assert len(result.samples) == shots


def test_HomodyneMeasurement_one_mode():
    shots = 20

    simulator = pq.PureFockSimulator(
        d=1, config=pq.Config(cutoff=20, seed_sequence=123, hbar=1)
    )

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=1.0)

        pq.Q(0) | pq.HomodyneMeasurement()

    result = simulator.execute(program, shots)

    assert len(result.samples) == shots

    assert np.allclose(
        result.samples,
        [
            [1.74960435],
            [0.27660972],
            [0.86902385],
            [0.77861923],
            [0.755815],
            [2.04047249],
            [2.42392663],
            [0.99484774],
            [2.06082907],
            [2.28110812],
            [1.43722324],
            [0.92598922],
            [2.07300598],
            [0.85314593],
            [1.87234924],
            [1.64877767],
            [2.4443],
            [0.89617885],
            [2.00713962],
            [1.44643782],
        ],
    )


def test_HomodyneMeasurement_two_modes():
    shots = 20

    simulator = pq.PureFockSimulator(
        d=2, config=pq.Config(cutoff=7, seed_sequence=123, hbar=1)
    )

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=0.5)
        pq.Q(1) | pq.Displacement(r=-0.5)

        pq.Q(0, 1) | pq.HomodyneMeasurement()

    result = simulator.execute(program, shots)

    assert len(result.samples) == shots

    assert np.allclose(
        result.samples,
        [
            [1.04248244, -1.84398138],
            [0.16191613, -1.34935628],
            [0.04870163, -0.14112144],
            [1.7169028, -1.16870184],
            [1.35378134, 0.15965135],
            [0.73006562, -1.16096225],
            [1.36596076, -1.2949428],
            [1.16525672, -0.45997211],
            [1.73727326, -1.26577442],
            [1.30008007, -0.68581022],
            [0.18825746, -1.39577179],
            [0.70315301, -0.51975498],
            [0.0714127, -2.24813514],
            [0.65586351, -0.24068384],
            [1.69412448, -0.52716291],
            [1.68723419, 0.02688775],
            [0.15661024, 0.02666197],
            [1.14206796, -1.12222433],
            [1.29485143, 0.08064955],
            [0.33514897, -0.65509742],
        ],
    )


def test_HomodyneMeasurement_two_modes_with_1_mode_sampled():
    shots = 20

    simulator = pq.PureFockSimulator(
        d=2, config=pq.Config(cutoff=7, seed_sequence=123, hbar=1)
    )

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=0.5)
        pq.Q(1) | pq.Displacement(r=-0.5)

        pq.Q(0) | pq.HomodyneMeasurement()

    result = simulator.execute(program, shots)

    assert len(result.samples) == shots

    assert np.allclose(
        result.samples,
        [
            [1.04248244],
            [-0.43057477],
            [0.16191613],
            [0.07150249],
            [0.04870163],
            [1.33342043],
            [1.7169028],
            [0.28775002],
            [1.35378134],
            [1.57409005],
            [0.73006562],
            [0.21888906],
            [1.36596076],
            [0.14603569],
            [1.16525672],
            [0.94163604],
            [1.73727326],
            [0.1890752],
            [1.30008007],
            [0.73928001],
        ],
    )


def test_HomodyneMeasurement_different_hbar_values():
    shots = 20

    simulator_hbar_2 = pq.PureFockSimulator(
        d=3, config=pq.Config(cutoff=7, seed_sequence=123, hbar=2)
    )
    simulator_hbar_3 = pq.PureFockSimulator(
        d=3, config=pq.Config(cutoff=7, seed_sequence=123, hbar=3)
    )

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=0.5)
        pq.Q(1) | pq.Displacement(r=-0.5)
        pq.Q(2) | pq.Squeezing(r=0.1)
        pq.Q(0, 1) | pq.Beamsplitter5050()
        pq.Q(1, 2) | pq.Beamsplitter5050()

        pq.Q(0, 2) | pq.HomodyneMeasurement()

    samples_hbar_2 = simulator_hbar_2.execute(program, shots).samples
    samples_hbar_3 = simulator_hbar_3.execute(program, shots).samples

    assert np.allclose(samples_hbar_2 / np.sqrt(2), samples_hbar_3 / np.sqrt(3))



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

import piquasso as pq


def test_from_fock_state_preserves_fock_probabilities():
    with pq.Program() as pure_state_preparation_program:
        pq.Q(1) | pq.StateVector([1])

    pure_simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=4))
    mixed_simulator = pq.FockSimulator(d=2, config=pq.Config(cutoff=4))
    pure_state_preparation_state = pure_simulator.execute(
        pure_state_preparation_program
    ).state

    beamsplitter = pq.Beamsplitter(theta=np.pi / 4, phi=np.pi / 3)

    with pq.Program() as pure_state_program:
        pq.Q(0, 1) | beamsplitter

    with pq.Program() as mixed_state_program:
        pq.Q(0, 1) | beamsplitter

    pure_state = pure_simulator.execute(
        pure_state_program,
        initial_state=pure_state_preparation_state,
    ).state

    mixed_state = mixed_simulator.execute(
        mixed_state_program,
        initial_state=pq.FockState.from_fock_state(pure_state_preparation_state),
    ).state

    pure_state_fock_probabilities = pure_state.fock_probabilities
    mixed_state_fock_probabilities = mixed_state.fock_probabilities

    assert np.allclose(mixed_state_fock_probabilities, pure_state_fock_probabilities)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numba as nb

import numpy as np

from functools import lru_cache

from piquasso._math.indices import get_auxiliary_modes

from .._utils import (
    get_fock_space_index,
    get_cutoff_fock_space_dimension,
    next_second_quantized,
)


@lru_cache()
@nb.njit(cache=True)
def calculate_indices_for_controlled_phase(d, cutoff, modes):
    int_dtype = np.int64

    full_occ_number = np.empty(d, dtype=int_dtype)

    aux_modes = get_auxiliary_modes(d, modes)

    full_occ_number[modes[0]] = 1
    full_occ_number[modes[1]] = 1

    size = get_cutoff_fock_space_dimension(d - 2, cutoff)

    indices = np.empty(size, dtype=int_dtype)

    aux_occ_number = np.zeros(d - 2, dtype=int_dtype)

    for i in range(size):
        full_occ_number[aux_modes] = aux_occ_number

        index = get_fock_space_index(full_occ_number)

        indices[i] = index

        aux_occ_number = next_second_quantized(aux_occ_number)

    return indices


@nb.njit(cache=True)
def calculate_indices_for_ising_XX(d, cutoff, modes):
    int_dtype = np.int64

    full_occ_number = np.empty(d, dtype=int_dtype)

    aux_modes = get_auxiliary_modes(d, modes)

    size = get_cutoff_fock_space_dimension(d - 2, cutoff)

    indices = np.empty((size, 4), dtype=int_dtype)

    aux_occ_number = np.zeros(d - 2, dtype=int_dtype)

    for i in range(size):
        full_occ_number[aux_modes] = aux_occ_number

        for j in range(4):
            full_occ_number[modes[0]] = j // 2
            full_occ_number[modes[1]] = j % 2

            index = get_fock_space_index(full_occ_number)

            indices[i, j] = index

        aux_occ_number = next_second_quantized(aux_occ_number)

    return indices



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np

from unittest.mock import patch

import piquasso as pq
from piquasso.api.connector import BaseConnector


@pytest.fixture
def FakePreparation():
    class FakePreparation(pq.Preparation):
        pass

    return FakePreparation


@pytest.fixture
def FakeGate():
    class FakeGate(pq.Gate):
        def __init__(self, **params):
            super().__init__(params=params)

    return FakeGate


@pytest.fixture
def FakeMeasurement():
    class FakeMeasurement(pq.Measurement):
        pass

    return FakeMeasurement


@pytest.fixture
def FakeState():
    class FakeState(pq.State):
        def __init__(
            self, d: int, connector: BaseConnector, config: pq.Config = None
        ) -> None:
            super().__init__(connector=connector, config=config)

            self._d = d

        def get_particle_detection_probability(
            self, occupation_number: np.ndarray
        ) -> float:
            """
            NOTE: This needs to be here to be able to instantiate this class.
            """
            raise NotImplementedError

        @property
        def d(self):
            return self._d

        @property
        def fock_probabilities(self):
            return list()

        def validate(self):
            pass

    return FakeState


@pytest.fixture
def FakeConfig():
    class FakeConfig(pq.api.config.Config):
        pass

    return FakeConfig


@pytest.fixture
def FakeConnector():
    class FakeConnector(pq.api.connector.BaseConnector):
        pass

    p = patch.multiple(FakeConnector, __abstractmethods__=set())
    p.start()
    yield FakeConnector
    p.stop()


@pytest.fixture
def FakeSimulator(
    FakeState, FakeConfig, FakePreparation, FakeGate, FakeMeasurement, FakeConnector
):
    def fake_calculation(state: FakeState, instruction: pq.Instruction, shots: int):
        return pq.api.result.Result(state=state)

    class FakeSimulator(pq.Simulator):
        _state_class = FakeState

        _config_class = FakeConfig

        _instruction_map = {
            FakePreparation: fake_calculation,
            FakeGate: fake_calculation,
            FakeMeasurement: fake_calculation,
        }

        _default_connector_class = FakeConnector

    return FakeSimulator



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from typing import Optional

import numpy as np

from piquasso.api.config import Config
from piquasso.api.exceptions import InvalidState
from piquasso.api.connector import BaseConnector

from piquasso._math.linalg import vector_absolute_square

from .state import PureFockState


class BatchPureFockState(PureFockState):
    r"""A simulated batch pure Fock state, containing multiple state vectors."""

    def __init__(
        self, *, d: int, connector: BaseConnector, config: Optional[Config] = None
    ) -> None:
        """
        Args:
            d (int): The number of modes.
            connector (BaseConnector): Instance containing calculation functions.
            config (Config): Instance containing constants for the simulation.
        """

        super().__init__(d=d, connector=connector, config=config)

    def _apply_separate_state_vectors(self, state_vectors):
        self.state_vector = self._np.array(
            state_vectors, dtype=self._config.complex_dtype
        ).T

    @property
    def _batch_size(self):
        return self.state_vector.shape[1]

    @property
    def _batch_state_vectors(self):
        for index in range(self._batch_size):
            yield self.state_vector[:, index]

    @property
    def nonzero_elements(self):
        return [
            self._nonzero_elements_for_single_state_vector(state_vector)
            for state_vector in self._batch_state_vectors
        ]

    @property
    def fock_amplitudes_map(self):
        maps = []
        for state_vector in self._batch_state_vectors:
            amplitude_map = defaultdict(complex)
            nonzero_elements = self._nonzero_elements_for_single_state_vector(
                state_vector
            )
            for elem in nonzero_elements:
                amplitude_map[elem[1]] = elem[0]
            maps.append(amplitude_map)
        return maps

    def __str__(self) -> str:
        partial_strings = []
        for partial_nonzero_elements in self.nonzero_elements:
            partial_strings.append(
                self._get_repr_for_single_state_vector(partial_nonzero_elements)
            )

        return "\n".join(partial_strings)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BatchPureFockState):
            return False
        return self._np.allclose(self.state_vector, other.state_vector)

    @property
    def fock_probabilities(self):
        return [
            vector_absolute_square(state_vector, self._connector)
            for state_vector in self._batch_state_vectors
        ]

    @property
    def norm(self):
        return [
            self._connector.np.sum(partial_fock_probabilities)
            for partial_fock_probabilities in self.fock_probabilities
        ]

    def normalize(self) -> None:
        norms = self.norm

        if self._config.validate and any(np.isclose(norm, 0) for norm in norms):
            raise InvalidState("The norm of a state in the batch is 0.")

        self.state_vector = self.state_vector / self._np.sqrt(norms)

    def validate(self) -> None:
        if not self._config.validate:
            return

        if not all(np.isclose(norm, 1.0) for norm in self.norm):
            raise InvalidState(
                "The sum of probabilities is not close to 1.0 for at least one state "
                "in the batch."
            )

    def mean_position(self, mode: int) -> np.ndarray:
        np = self._connector.np
        fallback_np = self._connector.fallback_np
        multipliers, left_indices, right_indices = self._get_mean_position_indices(mode)

        lhs = (multipliers * self.state_vector[left_indices].T).T
        rhs = self.state_vector[right_indices]

        return np.real(
            np.einsum("ij,ij->j", lhs, rhs) * fallback_np.sqrt(self._config.hbar / 2)
        )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Benchmarking the Piquasso vs. TheWalrus hafnian with repetitions.
"""

import time

import json

from piquasso._math.permanent import permanent

import matplotlib.pyplot as plt

import numpy as np

np.set_printoptions(suppress=True, linewidth=200)


if __name__ == "__main__":
    x = []
    y = []

    ITER = 100

    FILENAME = f"permanent_benchmark_{int(time.time())}.json"

    # Compilation
    d = 2
    A = np.random.rand(d, d) + 1j * np.random.rand(d, d)
    A = A + A.T
    occupation_numbers = 2 * np.ones(d, dtype=int)
    permanent(A, occupation_numbers, occupation_numbers)
    ###

    for d in range(2, 20 + 1):
        print(d)
        x.append(d)
        A = np.random.rand(d, d) + 1j * np.random.rand(d, d)

        A = A + A.T

        occupation_numbers = 2 * np.ones(d, dtype=int)
        sum_ = 0.0

        for _ in range(ITER):
            print("|", end="", flush=True)
            start_time = time.time()
            permanent(A, occupation_numbers, occupation_numbers)
            sum_ += time.time() - start_time

        y.append(sum_ / ITER)
        print()

        with open(FILENAME, "w") as f:
            json.dump(dict(x=x, y=y), f, indent=4)

    plt.scatter(x, y, label="Piquasso")
    plt.legend()
    plt.yscale("log")
    plt.xlabel("d [-]")
    plt.ylabel("Execution time [s]")

    plt.show()
